-- Sergio 'Sopze' del Pino @ 2024 -- The Junker dynamic engine props

local M = {}
M.type = "sopzeDynamicProp"
M.relevantDevice = nil

local powertrain = require("powertrain")

local junkerEngine = {}

local updateCount = 0
local defaultFPS= 8
local fastFPS= 16
local updateFPS = defaultFPS
local stepUpdate = 1 / updateFPS

local gaugesCount = 0
local gaugesFPS = 15
local stepGauges = 1 / gaugesFPS

local isEven= false

local lastThrottle= 0
local lastOutputAV= 0

local oilPressureSmoother = newTemporalSigmoidSmoothing(4, 20, 12, 2, 0)
local fuelPressureSmoother = newTemporalSigmoidSmoothing(3, 30, 19, 3, 0)

local function clamp(num, min, max)
  return math.max(math.min(max, num), min)
end

local function clamp01(num)
  return clamp(num, 0.0, 1.0)
end

local function init()

  local jbeamData = v.data["junkerEngineDetails"]

  if jbeamData then
    
    local device = powertrain.getDevice(jbeamData.inputName or v.data.powertrain[0].name or "mainEngine")
    
    if not device then return end

    local engine = {}
    local filter_right = {}
    local filter_left = {}

    local airFilterBreakTriggerBeam

    local jbeam= {
      engineShakeBeam= jbeamData.engineShakeBeam or {},
      airFilterShakeBeamLeft= jbeamData.airFilterShakeBeamLeft or {},
      airFilterShakeBeamRight= jbeamData.airFilterShakeBeamRight or {},
      airFilterBreakTriggerBeam= jbeamData.airFilterBreakTriggerBeam
    }

    for _, beam in pairs(v.data.beams) do
      if beam.name and beam.name ~="" then
        if jbeam.engineShakeBeam == beam.name then
          table.insert(engine, beam)
        elseif jbeam.airFilterShakeBeamLeft == beam.name then
          table.insert(filter_left, beam)
        elseif jbeam.airFilterShakeBeamRight == beam.name then
          table.insert(filter_right, beam)
        elseif jbeam.airFilterBreakTriggerBeam == beam.name then
          airFilterBreakTriggerBeam = beam
        end
      end
    end

    local rpmToAV = 0.104719755

    junkerEngine = {
      device= device,
      runningShake= jbeamData.runningShake or 1,
      ignitionShake= jbeamData.ignitionShake or 1,
      moaningFactor= jbeamData.moaningFactor or 1,
      idleAV= device.idleRPM*rpmToAV,
      ignitionShakeAV= device.idleRPM*rpmToAV*0.85,
      failureRangeAV= {1500*rpmToAV, 2200*rpmToAV, 3100*rpmToAV, 1.5/(700*rpmToAV), 1.25/(900*rpmToAV)},
      airFilterBroken= false,
      beams= {
        airFilterBreakTrigger= airFilterBreakTriggerBeam,
        engineShake= engine,
        filterShake= { r= filter_right, l=filter_left }
      }
    }

    -- using junker as prefix to prevent conflicts in case they add those measurements to the base game
    electrics.values.junker_oilpressure= 0
    electrics.values.junker_fuelpressure= 0

    --print("----------------------")
    --for match in string.gmatch(serialize(device),'[^,]+') do
    --  print(match)
    --end
  end
end

local function setUpdateRate(fps)
  updateFPS = fps or defaultFPS
  stepUpdate = 1 / updateFPS
end

local function reset()
  junkerEngine.airFilterBroken= false
  lastThrottle = 0
  setUpdateRate(defaultFPS)
end

local function onBeamBroke()
  junkerEngine.airFilterBroken= true
  --gui.message("Air filter broken", 4, "")
end

local function updateAirFilter()

  if not junkerEngine.airFilterBroken and junkerEngine.beams.airFilterBreakTrigger ~= nil then

    if obj:beamIsBroken(junkerEngine.beams.airFilterBreakTrigger.cid) then
      onBeamBroke()
    elseif junkerEngine.device and junkerEngine.device.outputAV1 > 10 then

      local force= 0

      if junkerEngine.device.outputAV1 < junkerEngine.ignitionShakeAV then
        force= (math.random()*1100 + junkerEngine.device.outputAV1*1.2) * junkerEngine.ignitionShake
      else
        force= (350*electrics.values.throttle+math.random()*250+150) * junkerEngine.runningShake
      end
      if math.random() > 0.5 then
        for _, v in pairs(junkerEngine.beams.filterShake.r) do
          obj:applyForce(v.id2, v.id1, force)
        end
      else
        for _, v in pairs(junkerEngine.beams.filterShake.l) do
          obj:applyForce(v.id2, v.id1, force)
        end
      end
    end
  end
end

local function getMoanFactor(av)
  if av < junkerEngine.failureRangeAV[2] then
    return clamp01((av-junkerEngine.failureRangeAV[1]) * junkerEngine.failureRangeAV[4])
  else
    return clamp01((junkerEngine.failureRangeAV[3] - av) * junkerEngine.failureRangeAV[5])
  end
end

local function updateEngine()
  if junkerEngine.device and junkerEngine.device.outputAV1 > 10 then
    
    local force= 0

    if junkerEngine.device.outputAV1 < junkerEngine.ignitionShakeAV then

      if updateFPS ~= fastFPS then setUpdateRate(fastFPS) end

      -- ignition shake
      force= (30000 + math.random()*14000 + 4000 * junkerEngine.device.outputAV1) * junkerEngine.ignitionShake
      if math.random() > 0.5 then
        for _, v in pairs(junkerEngine.beams.engineShake) do
          obj:applyForce(v.id1, v.id2, force)
        end
      else
        for _, v in pairs(junkerEngine.beams.engineShake) do
          obj:applyForce(v.id1, v.id2, -force)
        end
      end
    else

      if updateFPS ~= defaultFPS then setUpdateRate(defaultFPS) end
      
      -- engine moaning
      if junkerEngine.device.outputAV1 > junkerEngine.failureRangeAV[1] and junkerEngine.device.outputAV1 < junkerEngine.failureRangeAV[3] then
        
        local av= junkerEngine.device.outputAV1
        local moanFactor= getMoanFactor(av) * .78 * junkerEngine.moaningFactor

        if lastOutputAV > av then
          moanFactor= moanFactor * .125
        end

        local softFactor= 1.0 - clamp01(((av-lastOutputAV) * 0.0325))
        local value= clamp01(1.0 - moanFactor * softFactor)

        junkerEngine.device.outputAV1= av * value
      end

      -- running shake
      if electrics.values.throttle - 0.16 >= lastThrottle then
        force= (30000 + math.random()*6000 + 30000 * electrics.values.throttle) * junkerEngine.runningShake
        for _, v in pairs(junkerEngine.beams.engineShake) do
          obj:applyForce(v.id1, v.id2, force)
        end
      end

      for _, v in pairs(junkerEngine.beams.engineShake) do
        obj:applyForce(v.id1, v.id2, -40000*junkerEngine.device.engineLoad * junkerEngine.runningShake)
      end

      if math.random() < 0.35 then
        local value = math.random() * junkerEngine.runningShake
        for _, v in pairs(junkerEngine.beams.engineShake) do
          obj:applyForce(v.id1, v.id2, 20000 + 10000 * value)
        end
      end
    end
  end
end

-- fake measurement, just for visual purposes (i didnt wanted them to be static)
local function updatePressureGauges(dt)

  local fuelraw= 0
  local oilraw= 0

  if electrics.values.rpm > 50 then

    local running= junkerEngine.device.outputAV1 > 25
    local tEnv = obj:getEnvTemperature() - 273.15

    fuelraw= (running and 18 or 6) + electrics.values.throttle * 1.5 + tEnv * .08
    oilraw= (running and 35 or 27) - electrics.values.oiltemp * .2 + tEnv * .08 + electrics.values.rpm * .0135
  end

  electrics.values.junker_fuelpressure = fuelPressureSmoother:get(fuelraw, electrics.values.junker_fuelpressure >= 12 and dt or dt*4)
  electrics.values.junker_oilpressure = oilPressureSmoother:get(oilraw, electrics.values.junker_oilpressure >= 35 and dt or dt*4)
end

local function update(dt)

  updateCount = updateCount + dt
  if updateCount > stepUpdate then

    if electrics.values.ignitionLevel> 1 then
      updateAirFilter()
      updateEngine()
    end
    
    lastThrottle= electrics.values.throttle
    lastOutputAV= junkerEngine.device.outputAV1
    isEven = not isEven
    
    updateCount = 0
  end

  gaugesCount = gaugesCount + dt
  if gaugesCount > stepGauges then

    updatePressureGauges(dt)

    gaugesCount= 0
  end
  
end

M.init      = init
M.onReset   = reset
M.updateGFX = update

return M